@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import base64
 import requests
+import re
 from typing import Union, Optional
 from wxcadm import log
 from .common import *
@@ -16,6 +17,8 @@ from .person import UserGroups, Person
 from .applications import WebexApplications
 from .workspace import Workspace, WorkspaceLocation
 from .call_routing import CallRouting
+from .reports import Reports
+from .calls import Calls
 
 
 class Org:
@@ -23,8 +26,8 @@ class Org:
                  name: str,
                  id: str,
                  parent: Webex = None,
-                 people: bool = True,
-                 locations: bool = True,
+                 people: bool = False,
+                 locations: bool = False,
                  hunt_groups: bool = False,
                  call_queues: bool = False,
                  xsi: bool = False,
@@ -67,7 +70,8 @@ class Org:
         """The XSI details for the Organization"""
         self._params: dict = {"orgId": self.id}
         self._licenses: Optional[list] = None
-        self.people: list = []
+        self._wxc_licenses: Optional[list] = None
+        self._people: list = []
         '''A list of all of the Person instances for the Organization'''
         self.workspaces: Optional[list] = None
         """A list of the Workspace instances for this Org."""
@@ -81,6 +85,9 @@ class Org:
         self._roles: Optional[dict] = None
 
         self.call_routing = CallRouting(self)
+        """ The :py:class:`CallRouting` instance for this Org """
+        self.reports = Reports(self)
+        """ The :py:class:`Reports` instance for this Org """
 
         # Set the Authorization header based on how the instance was built
         self._headers = parent.headers
@@ -91,7 +98,6 @@ class Org:
         # Create a CSDM instance for CSDM work
         self._csdm = CSDM(self, self._parent._access_token)
 
-        # Get all the people if we aren't told not to
         if locations:
             self.get_locations()
         if xsi:
@@ -219,6 +225,12 @@ class Org:
         """ The :py:class:`WebexApplications` list with the :py:class:`WebexApplication` instances for this Org """
         return WebexApplications(parent=self)
 
+    @property
+    def calls(self):
+        """ The :py:class:`Calls` instance for this Org"""
+        return Calls(parent=self)
+
+
     def get_paging_group(self, id: str = None, name: str = None, spark_id: str = None):
         """ Get the PagingGroup instance associated with a given ID, Name, or Spark ID
 
@@ -351,6 +363,40 @@ class Org:
                     return num
         return None
 
+    def get_workspace_devices(self, workspace: Optional[Workspace] = None):
+        """ Get Webex Calling Workspaces and their associated Devices
+
+        When called without an argument, a list of Workspace instances is returned. Devices associated with each
+        Workspace can be accessed with the .devices attribute of each Workspace. If there are no Devices assigned
+        to a Workspace, .devices will return an empty list (i.e. []).
+
+        When called with a ``workspace`` argument, which is a Workspace instance, the Devices for that Workspace will
+        be returned as a list of Device instances.
+
+        Args:
+            workspace (Workspace, optional): A Workspace instance to show devices for
+
+        Returns:
+            list: Either a list of :py:class:`Workspace` or a list of :py:class:`Device`, depending on the argument
+
+        """
+        # TODO: For now, until the Workspaces API works for all WxC Workspaces, we use the .numbers as the key
+        if workspace is not None:
+            return workspace.devices
+        workspaces = []
+        for number in self.numbers:
+            if 'owner' in number.keys():
+                if isinstance(number['owner'], dict) and number['owner']['type'] == 'PLACE':
+                    print(number)
+                    ws_config = {'displayName': number['owner']['firstName']}
+                    workspace = Workspace(self, id=number['owner']['id'], config=ws_config)
+                    if number['owner']['lastName'] != '.':
+                        workspace.name += f" {number['owner']['lastName']}"
+                    workspace.location = number['location']
+                    workspaces.append(workspace)
+        return workspaces
+
+
     @property
     def devices(self):
         """All the Device instances for the Org
@@ -416,6 +462,14 @@ class Org:
                     return location
         return None
 
+    @property
+    def people(self):
+        """ A list of all of the Person instances for the Organization """
+        if not self._people:
+            return self.get_people()
+        else:
+            return self._people
+
     def get_person_by_id(self, id: str):
         """Get the Person instance associated with a given ID
 
@@ -431,14 +485,16 @@ class Org:
                 return person
         return None
 
-    def __get_wxc_licenses(self):
+    @property
+    def wxc_licenses(self):
         """Get only the Webex Calling licenses from the Org.licenses attribute
 
         Returns:
-            list[str]:
+            list[str]: The license IDs for each Webex Calling license
 
         """
-        log.info("__get_wxc_licenses started")
+        if self.licenses is None:
+            self.__get_licenses()
         license_list = []
         for license in self.licenses:
             if license['wxc_license']:
@@ -536,7 +592,7 @@ class Org:
         response = r.json()
         if r.status_code == 200:
             person = Person(response['id'], self, response)
-            self.people.append(person)
+            self._people.append(person)
             return person
         else:
             raise PutError(response['message'])
@@ -555,7 +611,7 @@ class Org:
         """
         success = webex_api_call("delete", f"v1/people/{person.id}")
         if success:
-            self.get_people()
+            self._people = []
             return True
         else:
             return False
@@ -609,10 +665,31 @@ class Org:
         self.locations.clear()
         api_resp = webex_api_call("get", "v1/locations", headers=self._headers, params=self._params)
         for location in api_resp:
-            this_location = Location(self, location['id'], location['name'], address=location['address'])
+            this_location = Location(self,
+                                     location['id'],
+                                     location['name'],
+                                     address=location['address'],
+                                     time_zone=location['timeZone'],
+                                     preferred_language=location['preferredLanguage'])
             self.locations.append(this_location)
         return self.locations
 
+    def create_location(self,
+                        name: str,
+                        time_zone: str,
+                        preferred_language: str,
+                        announcement_language: str,
+                        address: dict):
+        payload = {
+            'name': name,
+            'timeZone': time_zone,
+            'preferredLanguage': preferred_language,
+            'announcementLanguage': announcement_language,
+            'address': address
+        }
+        response = webex_api_call('post', '/v1/locations', params={'orgId': self.id}, payload=payload)
+        return response
+
     def get_workspaces(self):
         """Get the Workspaces and Workspace Locations for the Organizations.
 
@@ -783,7 +860,7 @@ class Org:
         return self.hunt_groups
 
     def get_people(self):
-        """Get all of the people within the Organization.
+        """ Get all people within the Organization.
 
         Also creates a Person instance and stores it in the Org.people attributes
 
@@ -799,16 +876,13 @@ class Org:
         people = webex_api_call("get", "v1/people", headers=self._headers, params=params)
         log.info(f"Found {len(people)} people.")
 
-        self.wxc_licenses = self.__get_wxc_licenses()
-
         for person in people:
             this_person = Person(person['id'], parent=self, config=person)
-            self.people.append(this_person)
-        return self.people
+            self._people.append(this_person)
+        return self._people
 
     def _get_person(self, match):
         log.info(f"Getting person: {match}")
-        self.wxc_licenses = self.__get_wxc_licenses()
         if "@" in match:
             params = {"max": "1000", "callingData": "true", "email": match, **self._params}
             url = "v1/people"
@@ -819,27 +893,9 @@ class Org:
             url = f"v1/people/{match}"
             response = webex_api_call("get", url, headers=self._headers, params=params)
             this_person = Person(response['id'], parent=self, config=response)
-        self.people.append(this_person)
+        self._people.append(this_person)
         return this_person
 
-    def get_wxc_people(self):
-        """Get all of the people within the Organization **who have Webex Calling**
-
-        .. deprecated:: 2.3.0
-            Use :meth:`wxc_people` instead
-
-        Returns:
-            list[Person]: List of Person instances of people who have a Webex Calling license
-
-        """
-        if not self.people:
-            self.get_people()
-        wxc_people = []
-        for person in self.people:
-            if person.wxc:
-                wxc_people.append(person)
-        return wxc_people
-
     @property
     def wxc_people(self):
         """Return all the people within the Organization **who have Webex Calling**
@@ -848,8 +904,6 @@ class Org:
             list[Person]: List of Person instances of people who have a Webex Calling license
 
         """
-        if not self.people:
-            self.get_people()
         wxc_people = []
         for person in self.people:
             if person.wxc:
@@ -885,3 +939,37 @@ class Org:
             if license['id'] == license_id:
                 return license['name']
         return None
+
+    def get_audit_events(self,
+                         start: str, end: str,
+                         actor: Optional[Person] = None) -> list[dict]:
+        """ Get a list of Admin Audit Events for the Organization
+
+        Args:
+            start (str): The first date/time in the report, in the format `YYYY-MM-DD` or `YYYY-MM-DDTHH:MM:SS.000Z`
+            end (str): The first date/time in the report, in the format `YYYY-MM-DD` or `YYYY-MM-DDTHH:MM:SS.000Z`
+            actor (Person, optional): Only show events performed by this Person
+
+        Returns:
+
+        """
+        # Normalize the start and end if only the date was provided
+        if re.fullmatch(r"\d{4}-\d{2}-\d{2}", start):
+            start = start + "T00.00.00.000Z"
+        if re.fullmatch(r"\d{4}-\d{2}-\d{2}", end):
+            end = end + "T23:59:59.999Z"
+
+        if actor is not None:
+            actor_id = actor.id
+        else:
+            actor_id = None
+
+        params = {
+            'orgId': self.id,
+            'from': start,
+            'to': end,
+            'actorId': actor_id,
+        }
+        response = webex_api_call('get', '/v1/adminAudit/events', params=params)
+        return response
+
