@@ -10,6 +10,7 @@ from collections import UserList
 
 from .common import *
 from .xsi import XSI
+from .device import Device
 
 from wxcadm import log
 
@@ -308,6 +309,18 @@ class Person:
         """
         return self._parent.usergroups.find_person_assignment(self)
 
+    @property
+    def devices(self):
+        log.info(f"Collecting devices for {self.display_name}")
+        devices = []
+        response = webex_api_call('get', f'/v1/telephony/config/people/{self.id}/devices')
+        log.debug(f"{response}")
+        for item in response['devices']:
+            this_device = Device(self, item)
+            devices.append(this_device)
+        return devices
+
+
 
     @property
     def hunt_groups(self):
@@ -553,6 +566,51 @@ class Person:
         else:
             return self.vm_config
 
+    def enable_vm_notification(self, email: str = None, push=True):
+        """
+        Change the Voicemail config to enable voicemail notification to specified email address. If the email param
+        is not present, it will use the Person's email address as the default.
+
+        Args:
+            email (optional): The email address to send VMs to.
+            push (optional): Whether to immediately push the change to Webex. Defaults to True.
+
+        Returns:
+            dict: The `Person.vm_config` attribute
+        """
+        if not self.vm_config:
+            self.get_vm_config()
+        if email is None:
+            email = self.email
+        self.vm_config['notifications']['enabled'] = True
+        self.vm_config['notifications']['destination'] = email
+        if push:
+            return self.push_vm_config()
+        else:
+            return self.vm_config
+
+    def disable_vm_notification(self, email: str = None, push=True):
+        """
+        Change the Voicemail config to disable voicemail notification to specified email address. If the email param
+        is not present, it will use the Person's email address as the default.
+
+        Args:
+            email (optional): The email address to send VMs to.
+            push (optional): Whether to immediately push the change to Webex. Defaults to True.
+
+        Returns:
+            dict: The `Person.vm_config` attribute
+        """
+        if not self.vm_config:
+            self.get_vm_config()
+        if email is None:
+            email = self.email
+        self.vm_config['notifications']['enabled'] = False
+        if push:
+            return self.push_vm_config()
+        else:
+            return self.vm_config
+
     def set_voicemail_rings(self, rings: int, push=True):
         """ Set the number of rings before an Unanswered call is sent to Voicemail
 
@@ -1137,6 +1195,20 @@ class Person:
             log.warning("The Executive Assistant config push failed")
             return False
 
+    @property
+    def wxc_numbers(self):
+        """ All the phone numbers for the user from the Webex Calling platform.
+
+        Unlike the :py:meth:`Person.numbers` property, which only reads the data from Webex Common Identity, this
+        property pulls the data from the Webex Calling platform, so it includes the Primary number as well as any
+        Alias numbers and doesn't include any Active Directory data.
+
+        Returns:
+            dict: Dict of numbers properties as defined at https://developer.webex.com/docs/api/v1/webex-calling-person-settings/get-a-list-of-phone-numbers-for-a-person
+
+        """
+        return webex_api_call('get', f'/v1/people/{self.id}/features/numbers')
+
 class Me(Person):
     """ The class representing the token owner. Some methods are only available at an owner scope. """
     def __init__(self, user_id, parent: Org = None, config: dict = None):
@@ -1263,7 +1335,7 @@ class UserGroups(UserList):
         groups = response['groups']
         log.debug(f"Webex returned {len(groups)} Groups")
         for group in groups:
-            usergroup = UserGroup(parent = self.parent, **group)
+            usergroup = UserGroup(parent=self.parent, **group)
             self.data.append(usergroup)
 
     def create_group(self, name: str,
