@@ -93,7 +93,7 @@ class XSIEventsChannelSet:
             channel = XSIEventsChannel(self, my_endpoint, my_events_endpoint)
             self.channels.append(channel)
 
-    def restart_failed_channel(self, channel: XSIEventsChannel):
+    def restart_failed_channel(self, channel: XSIEventsChannel, wait: int = 0):
         """ Starts a new channel, using an :py:class:`XSIEventsChannel` with an ``active=False`` status as the source
         to determine where the new channel should be established to.
 
@@ -104,11 +104,12 @@ class XSIEventsChannelSet:
             bool: True on success, False otherwise.
 
         """
-        log.debug(f"[{threading.current_thread().name}] ==== Restarting Failed XSIEventsChannel "
-                  f"with endpoint {channel.endpoint.split('/')[2]} ====")
         if channel.active is True:
             log.warning("Cannot restart an active channel")
             return False
+        log.debug(f"[{threading.current_thread().name}] ==== Restarting Failed XSIEventsChannel "
+                  f"with endpoint {channel.endpoint.split('/')[2]} in {wait} seconds ====")
+        time.sleep(wait)
         new_channel = XSIEventsChannel(self, channel.endpoint, channel.events_endpoint)
         self.channels.append(new_channel)
         # Send a DELETE on the failed channel, just to make sure it is removed from the server
@@ -131,6 +132,14 @@ class XSIEventsChannelSet:
 
         return True
 
+    @property
+    def active_channels(self):
+        active_channels = []
+        for channel in self.channels:
+            if channel.active is True:
+                active_channels.append(channel)
+        return active_channels
+
     def subscribe(self, event_package, person: Person = None):
         """ Subscribe to an Event Package over the channel opened with :meth:`XSIEvents.open_channel()`
 
@@ -143,11 +152,17 @@ class XSIEventsChannelSet:
             XSIEventsSubscription: The Subscription instance. False is returned if the subscription fails.
 
         """
+        log.info(f'Subscribing to {event_package}')
+        if len(self.active_channels) == 0:
+            log.warning("Cannot subscribe when no channels are active")
+            raise XSIError("Cannot subscribe with no active channels")
         subscription = XSIEventsSubscription(self, event_package, person=person)
-        if subscription:
+        if subscription and subscription.active is True:
             self.subscriptions.append(subscription)
+            log.debug(f'\tCurrent Subscriptions: {self.subscriptions}')
             return subscription
         else:
+            log.warning(f"The subscription did not activate")
             return False
 
     def unsubscribe(self, subscription_id: str):
@@ -181,6 +196,7 @@ class XSIEventsChannelSet:
         for subscription in self.subscriptions:
             subscription.delete()
         for channel in self.channels:
+            log.debug(f"Deleting Channel {channel.id} with {channel.endpoint}")
             channel.delete()
         return True
 
@@ -209,6 +225,7 @@ class XSIEventsChannel:
         self.events_endpoint = events_endpoint
         self.queue = self.parent.queue
         self.active = True
+        self.needs_restart = False
         self.last_refresh = ""
         self.xsp_ip:str = ''
         self.session = requests.Session()
@@ -245,18 +262,27 @@ class XSIEventsChannel:
 
         tid = tracking_id()
         log.debug(f"[{threading.current_thread().name}][{tid}] Sending Channel Start Request (Streaming) "
-                  f"to {self.endpoint.split('/')[2]}")
+                  f"to {self.endpoint}/v2.0/channel")
         log.debug(f"\tHeaders: {self._headers}\n\tPayload: {payload}")
         r = self.session.post(
             f"{self.endpoint}/v2.0/channel",
             headers={'TrackingID': tid, **self._headers}, data=payload, stream=True)
-        # Get the real IP of the remote server so we can send subsequent messages to that
-        ip, port = r.raw._connection.sock.getpeername()
+        # An issue with the F5 in Oct 2022 showed that we need to handle non-ok responses better
+        if not r.ok:
+            log.debug(f"\tHeaders: {r.headers}")
+            log.debug(f"\tBody: {r.text}")
+            log.warning(f"Unable to establish connection with {self.endpoint}")
+            ip = 'unknown'
+            self.active = False
+            self.needs_restart = True
+        else:
+            # Get the real IP of the remote server so we can send subsequent messages to that
+            ip, port = r.raw._connection.sock.getpeername()
         log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
                   f"Received response from {ip}. Status code {r.status_code}")
         # log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Headers: {r.headers}")
         log.debug(f"[{threading.current_thread().name}][{r.headers.get('TrackingID', 'Unknown')}] "
-                  f"Saving Cookies: {r.cookies}")
+                  f"Saving Cookies: {r.cookies.get_dict()}")
         self.cookies = r.cookies
 
         self.xsp_ip = str(ip)
@@ -316,6 +342,8 @@ class XSIEventsChannel:
                 chars = ""
         log.debug(f"[{threading.current_thread().name}][{tid}] Channel Loop ended: {self.id}")
         self.active = False
+        if self.needs_restart is True:
+            self.parent.restart_failed_channel(self, wait=60)
         return True
 
     def _heartbeat_daemon(self):
@@ -452,7 +480,7 @@ class XSIEventsSubscription:
             person (Person): A Person instance to subscribe to the event package for. If not provided,
                 the entire Organization will be subscribed.
         """
-        self.id = ""
+        self.id: Optional[str] = None
         self.parent = parent
         if person is None:
             self.target = "serviceprovider"
@@ -463,7 +491,7 @@ class XSIEventsSubscription:
         self.last_refresh = time.time()
         """ The date and time that the subscription was last refreshed """
         self.event_package = event_package
-        log.info(f"Subscribing to: {self.event_package}")
+        log.info(f"Subscribing to: {self.event_package} at {self.events_endpoint}")
         payload = '<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n' \
                   '<Subscription xmlns=\"http://schema.broadsoft.com/xsi\">' \
                   f'<event>{self.event_package}</event>' \
@@ -472,6 +500,11 @@ class XSIEventsSubscription:
         if person is None:
             r = requests.post(self.events_endpoint + f"/v2.0/serviceprovider/{self.parent.parent.enterprise}",
                               headers={"TrackingId": tracking_id(), **self._headers}, data=payload)
+            # Adding the following debug to help with failed subscriptions.
+            log.debug("Subscribe API Call:")
+            log.debug(f"\tURL: {self.events_endpoint}/v2.0/serviceprovider/{self.parent.parent.enterprise}")
+            log.debug(f"\tHeaders: {self._headers}")
+            log.debug(f"\tPayload: {payload}")
         else:
             if isinstance(person, wxcadm.person.Person):
                 # Make sure we have the Person's XSI Profile
@@ -482,8 +515,13 @@ class XSIEventsSubscription:
                 raise ValueError("The user argument requires a Person instance")
         if r.ok:
             response_dict = xmltodict.parse(r.text)
+            log.debug(f"Response: {response_dict}")
             self.id = response_dict['Subscription']['subscriptionId']
             log.debug(f"[{r.headers.get('TrackingID', 'Unknown')}] Subscription ID: {self.id}")
+            self.active = True
+        else:
+            log.warning(f"Subscription failed: [{r.status_code}] {r.text}")
+            self.active = False
 
     def _refresh_subscription(self):
         """ Refresh the subscription
@@ -620,21 +658,24 @@ class XSI:
         if get_profile:
             self._profile = self.profile
 
-    def new_call(self, address: str = None):
+    def new_call(self, address: str = None, phone: str = 'All', aor: Optional[str] = None):
         """Create a new Call instance
 
         Args:
             address (str, optional): The address to originate a call to
+            phone (str, optional): The phone to use for the call. Valid values are ``All``, ``Primary``,
+                or ``SharedCallAppearance``. The SharedCallAppearance argument also requires an address-or-record
+                ``aor`` argument with the AoR of the Shared device.
+            aor (str, optional): The address-of-record of the Shared device to place the call from
 
         Returns:
             Call: The Call instance
 
         """
         # If we got an address, pass it to the new instance
+        call = Call(self)
         if address is not None:
-            call = Call(self, address=address)
-        else:
-            call = Call(self)
+            call.originate(address=address, phone=phone, aor=aor)
         self._calls.append(call)
         return call
 
@@ -724,10 +765,18 @@ class XSI:
     def __get_xsi_data(self, url, params: Optional[dict] = None):
         if params is not None:
             params = {**params, **self._params}
+        log.debug(f"Sending API Call: {self.xsi_endpoints['actions_endpoint'] + url}")
+        log.debug(f"\tHeaders: {self._headers}")
+        log.debug(f"\tParams: {params}")
         r = requests.get(self.xsi_endpoints['actions_endpoint'] + url, headers=self._headers, params=params)
+        log.debug(f"Response: {r.status_code}")
         if r.status_code == 200:
             try:
                 response = r.json()
+                # As of the addition of directory() in 3.1.0, we have to be prepared for paginated responses.
+                # There is a chance to add that here so that the invoking method doesn't have to care about it, but
+                # for now we are going to just return the dict we get and let the other method decide if it needs more
+                # records and send this method the right params
             except json.decoder.JSONDecodeError:
                 response = r.text
             return_data = response
@@ -785,23 +834,28 @@ class XSI:
     @property
     def profile(self):
         """The XSI Profile for this Person"""
+        log.debug("Getting XSI Profile")
         if not self._profile or not self._cache:
             profile_data: dict = \
                 self.__get_xsi_data(f"/v2.0/user/{self.id}/profile")
             # The following is a mapping of the raw XSI format to the profile attribute
-            self._profile['registrations_url'] = profile_data['Profile']['registrations']['$']
-            self._profile['schedule_url'] = profile_data['Profile']['scheduleList']['$']
-            self._profile['fac_url'] = profile_data['Profile']['fac']['$']
-            self._profile['country_code'] = profile_data['Profile']['countryCode']['$']
-            self._profile['user_id'] = profile_data['Profile']['details']['userId']['$']
-            self._profile['group_id'] = profile_data['Profile']['details']['groupId']['$']
-            self._profile['service_provider'] = profile_data['Profile']['details']['serviceProvider']['$']
-            # Not everyone has a number and/or extension, so we need to check to see if there are there
-            if "number" in profile_data['Profile']['details']:
-                self._profile['number'] = profile_data['Profile']['details']['number']['$']
-            if "extension" in profile_data['Profile']['details']:
-                self._profile['extension'] = profile_data['Profile']['details']['extension']['$']
-            self._profile['raw'] = profile_data
+            log.debug(f"User Profile: {profile_data}")
+            if profile_data:
+                self._profile['registrations_url'] = profile_data['Profile']['registrations']['$']
+                self._profile['schedule_url'] = profile_data['Profile']['scheduleList']['$']
+                self._profile['fac_url'] = profile_data['Profile']['fac']['$']
+                self._profile['country_code'] = profile_data['Profile']['countryCode']['$']
+                self._profile['user_id'] = profile_data['Profile']['details']['userId']['$']
+                self._profile['group_id'] = profile_data['Profile']['details']['groupId']['$']
+                self._profile['service_provider'] = profile_data['Profile']['details']['serviceProvider']['$']
+                # Not everyone has a number and/or extension, so we need to check to see if there are there
+                if "number" in profile_data['Profile']['details']:
+                    self._profile['number'] = profile_data['Profile']['details']['number']['$']
+                if "extension" in profile_data['Profile']['details']:
+                    self._profile['extension'] = profile_data['Profile']['details']['extension']['$']
+                self._profile['raw'] = profile_data
+            else:
+                self._profile = None
         return self._profile
 
     @property
@@ -817,7 +871,7 @@ class XSI:
     def get_fac(self):
         # If we don't have a FAC URL, go get it
         if "fac_url" not in self._profile:
-            self.profile()
+            self.profile
         r = requests.get(self.xsi_endpoints['actions_endpoint'] + self._profile['fac_url'],
                          headers=self._headers, params=self._params)
         response = r.json()
@@ -851,6 +905,133 @@ class XSI:
                 self.services[service['name']['$']] = True
         return self.services
 
+    def directory(self,
+                  type: Optional[str] = 'Enterprise',
+                  first_name: Optional[str] = None,
+                  last_name: Optional[str] = None,
+                  name: Optional[str] = None,
+                  user_id: Optional[str] = None,
+                  group_id: Optional[str] = None,
+                  number: Optional[str] = None,
+                  extension: Optional[str] = None,
+                  mobile_number: Optional[str] = None,
+                  department: Optional[str] = None,
+                  email: Optional[str] = None,
+                  any_match: Optional[bool] = False,) -> list[dict]:
+        """ Search the Webex Calling directories
+
+        When search filters are applied as arguments, the directory will be searched by those values. If the desire is
+        to match any one of the values (logical OR), the ``any_match`` parameter should also be set to True. Note that
+        each directory type supports its own filter criteria and not all criteria are available across all directories.
+        The following table shows the filter arguments that are available for each type.
+
+        Note that all directory filter arguments are case-sensitive within Webex Calling. If you want to perform a
+        case-insensitive search, append "/i" to the search string. For example, ``first_name='John/i'`` will match
+        "John", "john" and "JOHN".
+
+        .. list-table:: Directory Types
+            :widths: 25 75
+            :header-rows: 1
+
+            * - Directory Type
+              - Available Filters
+            * - Enterprise
+              - first_name, last_name, name, user_id, group_id, number, extension, mobile_number, department, email
+            * - Group
+              - first_name, last_name, name, user_id, group_id, number, extension, mobile_number, department, email
+            * - Personal
+              - name, number
+
+
+        Args:
+            type (str, Optional): The type of diectory to search. Valid values are ``Enterprise``,
+                ``Group``, and ``Personal``. Defaults to ``Enterprise``.
+            first_name (str, optional): The First Name field in the directory
+            last_name (str, optional): The Last Name field in the directory
+            name (str, optional): The "combined" name field that allows a search based on First Name and Last Name
+            user_id (str, optional): The User ID used on the Webex Calling call control platform. This is the User ID
+                that is displayed for the user across all of the XSI APIs
+            group_id (str, optional): The Group ID used on the Webex Calling call control platform. This ID is unique
+                to the Webex Calling Location and can be used to find all users at a specific Location.
+            number (str, optional): The Webex Calling phone number
+            extension (str, optional): The Webex Calling extension
+            mobile_number (str, optional): The user's mobile number, if populated by Directory Sync
+            department (str, optional): The user's department, if populated by Directory Sync
+            email (str, optional): The user's email address
+            any_match (bool, optional): When True, all arguments will be treated uniquely (OR) and results will be
+                returned if any argument matches. For exmaple, if the arguments
+                ``first_name=Joe,last_name=Smith,any_match=True`` are passed, the results will include Joe Smith, but
+                will also include Joe James and Lisa Smith. Defaults to ``False``.
+
+        Returns:
+            list[dict]: A list of matching records, represented as a dictionary. If there are no matches, an empty
+                list is returned. If an invalid directory type is requested, None is returned.
+
+        """
+        type = type.title()
+        log.info(f"Getting {type} directory")
+        log.debug(f"\tArgs: {locals()}")
+        if type.upper() in ['ENTERPRISE', 'GROUP']:
+            params = {
+                'firstName': first_name,
+                'lastName': last_name,
+                'name': name,
+                'userId': user_id,
+                'groupId': group_id,
+                'number': number,
+                'extension': extension,
+                'mobileNo': mobile_number,
+                'department': department,
+                'emailAddress': email,
+                'searchCriteriaModeOr': any_match,
+            }
+        elif type.upper() in ['PERSONAL']:
+            params = {
+                'name': name,
+                'number': number
+            }
+        else:
+            log.warning(f"Directory type {type} is not valid")
+            return None
+
+        # Set some vars to keep track of the number of records we get and how many more we expect
+        more_records = True
+        num_records = 0
+        next_index = 1
+        get_count = 50
+        return_records = []
+        while more_records is True:
+            response = self.__get_xsi_data(f'/v2.0/user/{self.id}/directories/{type}',
+                                           params={'start': next_index, 'results': get_count, **params})
+            num_records += int(response[type]['numberOfRecords']['$'])
+            total_records = int(response[type]['totalAvailableRecords']['$'])
+            log.debug(f"Received {num_records}/{total_records} records")
+            if int(response[type]['numberOfRecords']['$']) == 0: # No records returned
+                return return_records
+            if type.upper() in ['ENTERPRISE', 'GROUP']:
+                key = f"{type.lower()}Directory"
+                if isinstance(response[type][key]['directoryDetails'], dict):   # If we get a dict instead of a list
+                    return_records.append(response[type][key]['directoryDetails'])
+                else:
+                    for entry in response[type][key]['directoryDetails']:
+                        return_records.append(entry)
+            if type.upper() in ['PERSONAL']:
+                if isinstance(response[type]['entry'], dict):   # Single entry returned
+                    return_records.append(response[type]['entry'])
+                else:
+                    for entry in response[type]['entry']:
+                        return_records.append(entry)
+
+            # Check and see if we need to get more records
+            if num_records < total_records:
+                more_records = True
+                next_index += int(response[type]['numberOfRecords']['$'])
+            else:
+                more_records = False
+
+        return return_records
+
+
 
 class Call:
     """
@@ -890,6 +1071,9 @@ class Call:
         """The externalTrackingId used by XSI"""
         self._status: dict = {}
         """The status of the call"""
+        self.held: bool = False
+        """ Whether or not the call is on hold """
+        self._transfer_call = None
         if type(self._parent) is wxcadm.person.Person:
             # This is where we set things based on whether the parent is a Person
             self._url = _url_base
@@ -915,7 +1099,11 @@ class Call:
         if address:
             self.originate(address)
 
-    def originate(self, address: str, comment: str = "", executive: str = None):
+    def originate(self, address: str,
+                  comment: str = "",
+                  phone: str = 'All',
+                  aor: Optional[str] = None,
+                  executive: str = None):
         """Originate a call on behalf of the Person
 
         The optional ``executive`` argument takes an extension or phone number and allows the call to be placed
@@ -925,6 +1113,10 @@ class Call:
         Args:
             address (str): The address (usually a phone number) to originate the call to
             comment (str, optional): Text comment to attach to the call
+            phone (str, optional): The phone to use for the call. Valid values are ``All``, ``Primary``,
+                or ``SharedCallAppearance``. The SharedCallAppearance argument also requires an address-or-record
+                ``aor`` argument with the AoR of the Shared device.
+            aor (str, optional): The address-of-record of the Shared device to place the call from
             executive (str, optional): The phone number or extension of the Executive to place the call on behalf of
 
         Returns:
@@ -949,7 +1141,9 @@ class Call:
                 raise NotAllowed("Person is not allowed to place calls on behalf of this executive")
         else:
             log.info(f"Originating a call to {address} for {self._userid}")
-            params = {"address": address, "info": comment}
+            params = {"address": address, "info": comment, "location": phone}
+            if aor is not None:
+                params['locationAddress'] = aor
             r = requests.post(self._url + "/new", headers=self._headers, params=params)
             if r.status_code == 201:
                 response = r.json()
@@ -1024,24 +1218,36 @@ class Call:
         response = r.json()
         log.debug(f"Call Status response: {response}")
         if r.status_code == 200:
-            return_data = {
-                "network_call_id": response['Call']['networkCallId']['$'],
-                "personality": response['Call']['personality']['$'],
-                "state": response['Call']['state']['$'],
-                "remote_party": {
-                    "address": response['Call']['remoteParty']['address']['$'],
-                    "call_type": response['Call']['remoteParty']['callType']['$'],
-                },
-                "endpoint": {
-                    "type": response['Call']['endpoint']['@xsi1:type'],
-                    "AoR": response['Call']['endpoint']['addressOfRecord']['$']
-                },
-                "appearance": response['Call']['appearance']['$'],
-                "diversion_inhibited": response['Call']['diversionInhibited'],
-                "start_time": response['Call']['startTime']['$'],
-                "answer_time": response['Call']['answerTime']['$'],
-                "status_time": int(time.time())
-            }
+            return_data = {}
+            ## Added key checking 3.0.4
+            call_status = response['Call']
+            if 'networkCallId' in call_status:
+                return_data['network_call_id'] = call_status['networkCallId']['$']
+            if 'personality' in call_status:
+                return_data['personality'] = call_status['personality']['$']
+            if 'state' in call_status:
+                return_data['state'] = call_status['state']['$']
+            if 'remoteParty' in call_status:
+                return_data['remote_party'] = {
+                    "address": call_status['remoteParty']['address']['$'],
+                    "call_type": call_status['remoteParty']['callType']['$']
+                }
+            if 'endpoint' in call_status:
+                return_data['endpoint'] = {
+                    "type": call_status['endpoint']['@xsi1:type'],
+                    "aor": call_status['endpoint']['addressOfRecord']['$']
+                }
+            if 'appearance' in call_status:
+                return_data['appearance'] = call_status['appearance']['$']
+            if 'diversionInhibited' in call_status:
+                return_data['diversion_inhibited'] = call_status['diversionInhibited']
+            if 'startTime' in call_status:
+                return_data['start_time'] = call_status['startTime']['$']
+            if 'answerTime' in call_status:
+                return_data['answer_time'] = call_status['answerTime']['$']
+            else:
+                return_data['answer_time'] = None
+            return_data['status_time'] = int(time.time())
             return return_data
         else:
             return False
@@ -1093,7 +1299,7 @@ class Call:
         has been done first.
 
         Returns:
-            bool: Whether or not the transfer completes
+            bool: Whether the transfer completes
 
         """
         log.info("Completing transfer...")
@@ -1105,6 +1311,10 @@ class Call:
         Starts a multi-party conference. If the call is already held and an attended transfer is in progress,
         meaning the user is already talking to the transfer-to user, this method will bridge the calls.
 
+        When an address is passed, the existing call will be placed on hold, the second call will be originated, and the
+        conference will be connected as soon as the second party answers. If the desired behavior is not to connect the
+        parties until later, the calls should be created separately and bridged later.
+
         Args:
             address (str, optional): The address (usually a phone number or extension) to conference to. Not needed
                 when the call is already part of an Attended Transfer
@@ -1136,8 +1346,34 @@ class Call:
             else:
                 return False
         else:
-            # Still needs work.
-            pass
+            # Put the current call on hold if it isn't already
+            if self.held is False:
+                self.hold()
+            # Build the second call
+            second_call = self._parent.new_call(address=address)
+            # Get the status of the second call. We cannot complete the conference until it is answered.
+            call_state = second_call.status.get('state', 'unknown')
+            while call_state != 'Active':
+                call_state = second_call.status.get('state', 'unknown')
+
+            xml = f"<?xml version=\"1.0\" encoding=\"UTF-8\"?>" \
+                  f"<Conference xmlns=\"http://schema.broadsoft.com/xsi\">" \
+                  f"<conferenceParticipantList>" \
+                  f"<conferenceParticipant>" \
+                  f"<callId>{self.id}</callId>" \
+                  f"</conferenceParticipant>" \
+                  f"<conferenceParticipant>" \
+                  f"<callId>{second_call.id}</callId>" \
+                  f"</conferenceParticipant>" \
+                  f"</conferenceParticipantList>" \
+                  f"</Conference>"
+            headers = self._headers
+            headers['Content-Type'] = "application/xml; charset=UTF-8"
+            r = requests.post(self._url + f"/Conference", headers=headers, data=xml)
+            if r.status_code in [200, 201, 204]:
+                return self._parent.new_conference([self.id, second_call.id])
+            else:
+                return False
 
     def send_dtmf(self, dtmf: str):
         """Transmit DTMF tones outbound
@@ -1161,6 +1397,7 @@ class Call:
 
         """
         r = requests.put(self._url + f"/{self.id}/Hold", headers=self._headers)
+        self.held = True
         return XSIResponse(r)
 
     def resume(self):
@@ -1171,6 +1408,12 @@ class Call:
 
         """
         r = requests.put(self._url + f"/{self.id}/Talk", headers=self._headers)
+        self.held = False
+        return XSIResponse(r)
+
+    def answer(self):
+        r = requests.put(self._url + f"/{self.id}/Talk", headers=self._headers)
+        self.held = False
         return XSIResponse(r)
 
     def park(self, extension: str = None):
@@ -1215,6 +1458,7 @@ class Call:
     def reconnect(self):
         """Retrieves the call from hold **and releases all other calls**"""
         r = requests.put(self._url + f"{self.id}/Reconnect", headers=self._headers)
+        self.held = False
         return XSIResponse(r)
 
     def recording(self, action: str):
